#!/usr/bin/python2.5

"""Use contents manifest to improve rsync performance by only syncing
directories which are out-of-date.

Updates content manifests in-place if you use a split manifest.  Otherwise
requires both source and destinations contents manifests.
"""

from __future__ import with_statement
import os, os.path, shutil, stat, sys
from bitfrost.contents.utils import mkhashes, open_envel, UnifiedContents
from subprocess import check_call

def rsync (src, dst, path, recursive, verbosity=0):
    if verbosity > 0: print "Rsyncing: "+path
    # the '-c' option is a bit sketch; we'd like to use filetimes, but my
    # current archive mechanism (untar all files in a build) doesn't allow it.
    # Have to be careful that build/archive mechanism maintains timestamps if
    # we want to use them.
    args = ['/usr/bin/rsync', '-aOyz', '-l', '-c', '--delete',
            '--numeric-ids', '--no-whole-file',
            os.path.join(src, path)+'/', os.path.join(dst, path)]
    if not recursive:
        args[2:2] = ['--no-r', '--dirs']
    if verbosity > 1: args[2:2] = ['-v']
    check_call(args)

def sync_and_check (src, dst, path, old_contents, new_contents, verbosity=0):
    """Do an rsync on the dirty directory given by 'path'.  Afterwards,
    check the .rmanifest to see if any of the other directories need
    syncing, and if so, do it."""

    rsync(src, dst, path, False, verbosity)
    __, dirobject = new_contents.get_dirobject_pair(path)
    __, contents_map = open_envel('dir', 1, dirobject)
    for name, entry in contents_map.items():
        if stat.S_ISDIR(entry['m']):
            # in non-recursive mode, rsync creates the directory but doesn't
            # fill it.
            try:
                dm = old_contents[os.path.join(path, name)]
                if entry['h'] != mkhashes(dm):
                    raise RuntimeError("hash doesn't match")
            except:
                sync_and_check(src, dst, os.path.join(path, name), old_contents, new_contents, verbosity)

## Updaters built on the above primitives.

def _get_contents(url, contents):
    check_call(['rsync','-z',url+'/contents', contents])

def olpc_update_irsync_pristine(vhash, root, contents, tmpdir,
                                url, clean_cb, verbosity, **extra):
    """Do an incremental rsync, starting from a pristine tree."""
    if verbosity >=0: print >>sys.stderr, " - Cleaning tree."
    clean_cb() # want a clean tree
    if verbosity >=0: print >>sys.stderr, " - Fetching contents."
    new_contents = os.path.join(tmpdir, 'contents')
    shutil.copy(contents, new_contents)
    _get_contents(url, new_contents)
    old_contents_obj = UnifiedContents(contents)
    new_contents_obj = UnifiedContents(new_contents)
    old_contents_obj[''] # side-effect: load subdirs
    if verbosity >=0: print >>sys.stderr, " - Performing incremental rsync."
    sync_and_check(url+'/root', root, '', old_contents_obj, new_contents_obj,
                   verbosity)
    os.rename(new_contents, contents)
    # NASTY HACK: since we don't have COW to protect our pristine tree,
    #   /var/run/utmp and /var/log/wtmp are being stomped on.
    # We also don't protect /etc/alsa/asound.state =(
    for hack_dir in [ 'var', 'etc/alsa' ]:
        rsync(url+'/root', root, hack_dir, True, verbosity)

def olpc_update_irsync_dirty(vhash, root, contents, tmpdir,
                             url, clean_cb, verbosity, **extra):
    """Do an incremental rsync, not assuming that tree is pristine."""
    if verbosity >=0: print >>sys.stderr, " - Creating contents for existing tree."
    old_contents = os.path.join(tmpdir, 'contents')
    check_call(['/usr/sbin/olpc-contents-create','-f',old_contents,
                '-p','%s/etc/passwd'%root, '-g', '%s/etc/group'%root,
                root])
    if verbosity >=0: print >>sys.stderr, " - Fetching contents for update tree."
    _get_contents(url, contents)
    old_contents_obj = UnifiedContents(old_contents)
    new_contents_obj = UnifiedContents(contents)
    old_contents_obj[''] # side-effect: load subdirs
    if verbosity >=0: print >>sys.stderr, " - Performing incremental rsync."
    sync_and_check(url+'/root', root, '', old_contents_obj, new_contents_obj,
                   verbosity)
    os.unlink(old_contents)

def olpc_update_rsync(vhash, root, contents, tmpdir,
                      url, clean_cb, verbosity, **extra):
    """Do a full rsync."""
    if verbosity >=0: print >>sys.stderr, " - Fetching contents."
    _get_contents(url, contents)
    if verbosity >=0: print >>sys.stderr, " - Performing full rsync."
    rsync(url+'/root', root, '', True, verbosity)
