"""OLPC updater."""
from __future__ import with_statement
from tempfile import mkdtemp
from shutil import rmtree
from subprocess import call, check_call, Popen, PIPE
import os, os.path, re, struct, sys
from bitfrost.contents.utils import UnifiedContents

def inhibit_suspend(f):
    """Decorator to inhibit suspend during execution of the given function."""
    INHIBIT_FILE = '/etc/ohm/inhibit-suspend'
    def new_f(*args, **kwds):
        was_inhibited = os.path.exists(INHIBIT_FILE)
        if not was_inhibited: open(INHIBIT_FILE, 'w').close()
        try:
            return f(*args, **kwds)
        finally:
            if not was_inhibited: os.unlink(INHIBIT_FILE)
    new_f.func_name = f.func_name
    return new_f

def shorten_hash(vhash):
    """Return user-friendly short form of the version hash."""
    return vhash[0:32]

def rewrite_symlink(src, dst):
    """Atomically swing a symlink at dst from its current contents to src."""
    tmpdir = mkdtemp(prefix='sym', dir=os.path.dirname(dst))
    tmplnk = os.path.join(tmpdir, 'symlink')
    try:
        os.symlink(src, tmplnk)
        os.rename(tmplnk, dst)
    finally:
        rmtree(tmpdir, ignore_errors=True)

def trim_config(partitioned, cur_hash):
    """Remove 'alt' from current boot config."""
    if not partitioned:
        swing_config(partitioned, current=cur_hash, alt=None)
    else:
        #0-2. Make /boot/alt point to ../`basename /versions/running`
        assert False # XXX: WRITE ME

def swing_config(partitioned, current, alt=None):
    """Create new boot configuration."""
    if not partitioned:
        # 11. Make a new config /versions/configs/$d (d = mkdtemp)
        cfgdir = mkdtemp(prefix='cfg.', dir='/versions/configs/')
        # 12. Create 'current' symlink to ../../pristine/<hash>
        # XXX: should be ../../run/%s (trac #4898)
        os.symlink('/versions/pristine/%s' % shorten_hash(current),
                   os.path.join(cfgdir, 'current'))
        # 13. Create 'alt' symlink to *realpath of* /versions/running
        if alt is not None:
            # XXX: should be ../../run/%s (trac #4898)
            os.symlink('/versions/pristine/%s' % shorten_hash(alt),
                       os.path.join(cfgdir, 'alt'))
        # 14. Swing /versions/boot to /versions/configs/$d
        #      (atomic!  iff we do file move of new symlink)
        oldcfg = os.path.basename(os.readlink('/versions/boot').rstrip('/'))
        rewrite_symlink('configs/%s' % os.path.basename(cfgdir.rstrip('/')),
                        '/versions/boot')
        # 15. Delete /versions/configs/$c
        rmtree('/versions/configs/%s' % oldcfg)
    else:
        # 11. Copy /versions/pristine/<hash>/boot to boot:/boot-versions/<hash>
        # 12. Make boot:/boot-versions/<hash>/alt point to what boot:/boot currently points to
        # 13. Atomically swing boot:/boot to point to /boot-versions/<hash>
        assert False # XXX

def clean_versions(report, partitioned, cur_hash, new_hash):
    """Free up space for update by removing old or incomplete versions."""
    def rmrf(f, ignore_errors=False):
        if os.path.isdir(f) and not os.path.islink(f):
            rmtree(f, ignore_errors=ignore_errors)
        else:
            try:
                os.unlink(f)
            except:
                if not ignore_errors: raise
    # 3a. Clean up pristine and running trees.
    keep = [ shorten_hash(cur_hash) ]
    try:
        keep.extend(os.listdir('/versions/sticky'))
    except: pass # if /versions/sticky doesn't exist, don't worry about it.
    for f in os.listdir('/versions/pristine'):
        assert len(f) > 0 and '.' not in f # defensive programming.
        if f in keep: continue
        report(0, 'Deleting old pristine version %s' % f)
        rmrf('/versions/pristine/%s' % f, ignore_errors=True)
    for f in os.listdir('/versions/run'):
        assert len(f) > 0 and '.' not in f # defensive programming.
        if f in keep: continue
        report(0, 'Deleting old run version %s' % f)
        rmrf('/versions/run/%s' % f, ignore_errors=True)
    # 3b. Delete corresponding members of /boot-versions if using a boot partition
    if partitioned: assert False
    # 4. Clean up old configs.
    for f in os.listdir('/versions/configs'):
        assert len(f) > 0 and not f.startswith('.') # defensive programming.
        if f == os.path.basename(os.readlink('/versions/boot').rstrip('/')):
            continue
        report(0, 'Deleting old boot configuration %s' % f)
        rmrf('/versions/configs/%s' % f, ignore_errors=True)
    # 4b. Delete corresponding member of /boot-versions if using a boot partition.
    if partitioned: assert False
    # 5. Clean up updates directory.
    for f in os.listdir('/versions/updates'):
        assert len(f) > 0 and not f.startswith('.') # defensive programming.
        if f.startswith(shorten_hash(new_hash)): continue
        report(0, 'Deleting old incomplete update %s' % f)
        rmrf('/versions/updates/%s' % f, ignore_errors=True)

def find_pristine(path, outfile):
    inodes = sorted([int(line.strip()) for line in Popen
                     (['/usr/bin/find',path,'-type','f','-printf','%i\\n'],
                      stdout=PIPE).stdout.readlines()])
    outfile.write(struct.pack('=%dL' % len(inodes), *inodes))

def update_validates(new_hash):
    # check that hash of contents file is correct and that contents
    # file validates.
    updir = '/versions/updates/%s' % shorten_hash(new_hash)
    cmd = ['/usr/sbin/olpc-contents-verify','%s/etc/passwd' % updir,
           '%s/etc/group' % updir, new_hash, '%s.contents' % updir,
           updir]
    if os.path.exists(updir+'.pristine'):
        cmd.append(updir+'.pristine') # pristine inodes file
    check_call(cmd)
    return True # verification successful!

def reboot():
    """
    # CJB says something like this ought to work, but there's probably
    # sugar magic i need to perform.
    from hardware import hardwaremanager
    hw_manager = hardwaremanager.get_manager()
    hw_manager.reboot() # was .shutdown()
    # maybe I should wait a while and call 'shutdown' explicitly?
    """ # XXX see trac #4878
    check_call(['/sbin/shutdown','-r','now'])

@inhibit_suspend
def perform_update(new_hash, priority, hints, verbosity=0):
    """Do initial setup for updates, validate result trees, and install
    successful updates."""
    # see http://wiki.laptop.org/go/Early_Boot#Upgrade_procedure
    def report(lvl, msg):
        if verbosity>=lvl: print >>sys.stderr, msg
    # -1: Check that /versions/pristine/<hash> doesn't already exist.
    if re.match(r'[0-9a-f]+$', new_hash) is None:
        report(0, 'Bad hash: %s' % new_hash)
        sys.exit(1)
    if os.path.exists('/versions/pristine/%s' % shorten_hash(new_hash)):
        report(0, 'Given version %s already exists.' % shorten_hash(new_hash))
        sys.exit(1)
    cur_hash = os.path.basename(os.readlink('/versions/running'))
    partitioned = False # XXX: how do we tell?
    # steps 0-2: remove old version from boot config.
    trim_config(partitioned, cur_hash)
    # steps 3-5: clean up old or incomplete versions to make space.
    clean_versions(report, partitioned, cur_hash, new_hash)
    # 5b. If /versions/updates/<new hash> doesn't exist, create it.
    updir = '/versions/updates/%s' % shorten_hash(new_hash)
    upcon = updir+'.contents'
    uppri = updir+'.pristine'
    uptmp = mkdtemp(prefix='tmp.', dir='/versions/updates')
    def make_clean_update():
        rmtree(updir, ignore_errors=True)
        for f in [ upcon, uppri ]:
            try:
                os.unlink(f)
            except: pass # ignore errors
        check_call(['/bin/cp','-al',
                    '/versions/pristine/%s' % shorten_hash(cur_hash), updir])
        call(['/bin/cp','-al',
              '/versions/contents/%s' % shorten_hash(cur_hash), upcon])
        with open(uppri, 'wb') as f:
            find_pristine('/versions/pristine/%s' % shorten_hash(cur_hash), f)
    clean_cb = make_clean_update
    if not os.path.exists(updir):
        make_clean_update()
        clean_cb = lambda : None # already clean
    # 6-8: OLPC updater.
    #   loop through hints:
    #    make $updir into new updated version,
    #    leave a contents file matching new_hash in
    #       /versions/updates/<new hash>.contents
    import bitfrost.update # parent package has updaters list.
    updaters = vars(bitfrost.update) # all updaters defined in __init__
    for method, url in hints:
        report(0, "Trying %s update from %s" % (method, url))
        method = 'olpc_update_' + method # restrict namespace
        if method not in updaters: continue # unknown update method
        # invoke the update method.  should leave a contents file in place.
        try:
            # updaters should be tolerant of extra keyword arguments.
            updaters[method](vhash=new_hash, root=updir, contents=upcon,
                             tmpdir=uptmp, url=url,
                             clean_cb=clean_cb, verbosity=verbosity)
            # check that hash of contents file is correct and that contents
            # file validates.
            report(0, 'Verifying update.')
            if update_validates(new_hash):
                report(0, 'Update verified.')
                break # we're done!
        except:
            report(0, "Update failed: %s" % str(sys.exc_info()[1]))
            # go on to next method.
        clean_cb = make_clean_update # definitely dirty now.
    else:
        # update failed.
        report(0, "No more update methods to try.")
        sys.exit(2)
    # successful update!
    # 10. Move /versions/updates/<hash> to /versions/pristine/<hash>
    os.rename(updir, '/versions/pristine/%s' % shorten_hash(new_hash))
    os.rename(upcon, '/versions/contents/%s' % shorten_hash(new_hash))
    os.unlink(uppri)
    rmtree(uptmp, ignore_errors=True)
    #  10b. Create /versions/run/<hash> from /versions/pristine/<hash> according to P_SF_RUN setting
    check_call(['/bin/cp','-al',
                '/versions/pristine/%s' % shorten_hash(new_hash),
                '/versions/run/%s' % shorten_hash(new_hash)])
    # XXX: someday reimplement P_SF_RUN using COW/ro-bind-mounts, etc.
    # 11-15: make new boot configuration.
    swing_config(partitioned,
                 current=shorten_hash(new_hash), alt=shorten_hash(cur_hash))
    call(['/bin/sync']) # superstition
    #  16. If <priority> reboot.
    if priority=='urgent':
        report(0, "Rebooting now.")
        reboot()
    # Done!
    sys.exit(0) # success!
